using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Net;
using InfiniTec.Exchange.WebDav;
using InfiniTec.Threading;

namespace InfiniTec.Exchange
{
    class ExchangeStoreReference : EntryIdReference
    {
        private Guid _StoreId;
        private bool _IsLocalStore;
        private string _Path;
        private EmailAddressIdentifier _EmailIdentifier;
        private bool _IsDistributionList;

        public Guid StoreId
        {
            get { return _StoreId; }
        }

        public ExchangeStoreReference(byte[] buffer)
            : this(new BinaryReader(new MemoryStream(buffer)), 0, Guid.Empty, null)
        {
        }

        public ExchangeStoreReference(BinaryReader reader, Guid localStoreId, string url)
        {
            reader.ReadUInt32();

            switch (reader.ReadUInt32())
            {
                case 0:
                    _EmailIdentifier = EmailAddressIdentifier.EmailAddress1;
                    break;
                case 1:
                    _EmailIdentifier = EmailAddressIdentifier.EmailAddress2;
                    break;
                case 2:
                    _EmailIdentifier = EmailAddressIdentifier.EmailAddress3;
                    break;
                default:
                    _EmailIdentifier = EmailAddressIdentifier.None;
                    break;
            }

            ReadItemLocation(reader, url);

            _IsLocalStore = _StoreId == localStoreId;
        }

        public ExchangeStoreReference(BinaryReader reader, WindowsAddressBookEntryType entryType, Guid localStoreId, string url)
        {
            switch (entryType & WindowsAddressBookEntryType.EmailIndexMask)
            {
                case WindowsAddressBookEntryType.EmailIndex1:
                    _EmailIdentifier = EmailAddressIdentifier.EmailAddress1;
                    break;
                case WindowsAddressBookEntryType.EmailIndex2:
                    _EmailIdentifier = EmailAddressIdentifier.EmailAddress2;
                    break;
                case WindowsAddressBookEntryType.EmailIndex3:
                    _EmailIdentifier = EmailAddressIdentifier.EmailAddress3;
                    break;
                default:
                    _EmailIdentifier = EmailAddressIdentifier.None;
                    break;
            }

            _IsDistributionList = ((entryType & WindowsAddressBookEntryType.ContactDL) ==
                                   WindowsAddressBookEntryType.ContactDL);

            ReadItemLocation(reader, url);
            _IsLocalStore = _StoreId == localStoreId;
        }

        private void ReadItemLocation(BinaryReader reader, string baseUrl)
        {
            Guid folderId;
            ulong folderCnt;

            reader.ReadUInt32();
            _StoreId = new Guid(reader.ReadBytes(16));

            reader.ReadUInt16();

            folderId = ReadGuid(reader);
            folderCnt = SwapUInt64(reader.ReadUInt64());

            if (reader.BaseStream.Length - reader.BaseStream.Position >= 24)
            {
                Guid messageId;
                messageId = ReadGuid(reader);
                ulong messageCnt;
                messageCnt = SwapUInt64(reader.ReadUInt64());
                _Path =
                    string.Format(CultureInfo.CurrentCulture, baseUrl + "/-FlatUrlSpace-/{0:N}-{1:x}/{2:N}-{3:x}", folderId, folderCnt, messageId, messageCnt);
            }
            else
            {
                _Path = string.Format(CultureInfo.CurrentCulture, baseUrl + "/-FlatUrlSpace-/{0:N}-{1:x}", folderId, folderCnt);
            }

        }

        private static Guid ReadGuid(BinaryReader reader)
        {
            int a;
            short b, c;

            a = SwapInt(reader.ReadUInt32());
            b = reader.ReadInt16();
            c = SwapShort(reader.ReadUInt16());
            return new Guid(a, b, c, reader.ReadBytes(8));
        }

        private static short SwapShort(ushort value)
        {
            unchecked
            {
                ushort result;
                result = (ushort)(((value & 0xFF00) >> 8) |
                               ((value & 0x00FF) << 8));

                return (short)result;
            }
        }

        private static int SwapInt(uint value)
        {
            uint result;

            result = ((value & 0xFF000000) >> 24) |
                     ((value & 0x00FF0000) >> 8) |
                     ((value & 0x0000FF00) << 8) |
                     ((value & 0x000000FF) << 24);

            unchecked
            {
                return (int)result;
            }
        }


        private static ulong SwapUInt64(ulong value)
        {
            uint lo;
            uint hi;
            ulong result;

            lo = (uint)(value & 0xffffffff);
            hi = ((uint)(value >> 32)) & 0xffffffff;

            lo = ((lo & 0xFF000000) >> 8) |
                 ((lo & 0x00FF0000) << 8) |
                 ((lo & 0x0000FF00) >> 8) |
                 ((lo & 0x000000FF) << 8);

            hi = ((hi & 0xFF000000) >> 8) |
                 ((hi & 0x00FF0000) << 8) |
                 ((hi & 0x0000FF00) >> 8) |
                 ((hi & 0x000000FF) << 8);

            result = (((ulong)lo) << 32) | hi;

            return result;
        }

        protected override IEnumerable<OperationAction> ResolveInternal(Connection connection)
        {
            Item item;

            if (!_IsLocalStore)
            {
                yield break;
            }

            if (_IsDistributionList)
            {
                item = new DistributionList(_Path, connection);
                ((DistributionList)item).ResolveMembers = false;
            }
            else
            {
                item = new ContactItem(_Path, connection);
            }
            
            using (AsyncOperation operation = item.GetRefreshOperation())
            {
                yield return new NestedOperation(operation, false, typeof(WebException));

                if (operation.Status == OperationStatus.Completed)
                {
                    yield return new OperationResult<IAddressListEntry>((IAddressListEntry) item);
                }
                else
                {
                    WebException exception;

                    exception = operation.Exception as WebException;

                    if (exception == null ||
                        (((HttpWebResponse)exception.Response).StatusCode == HttpStatusCode.NotFound))
                    {
                        yield return new OperationResult<IAddressListEntry>(null);
                    }
                    else
                    {
                        throw exception;
                    }
                }
            }
        }
    }
}