using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using InfiniTec.DirectoryServices;
using InfiniTec.Threading;

namespace InfiniTec.Exchange
{
    /// <summary>
    /// Provides access to the properties of an Active Directory group
    /// </summary>
    public class Group: IAddressListEntry
    {
        /// <summary>
        /// This event is triggered, when a refresh operation completes.
        /// </summary>
        public event AsyncCompletedEventHandler RefreshCompleted
        {
            add { _BaseItem.RefreshCompleted += value; }
            remove { _BaseItem.RefreshCompleted -= value; }
        }

        private ActiveDirectoryGroup _BaseItem;

        /// <summary>
        /// Initializes a new instance of the <see cref="User"/> class.
        /// </summary>
        /// <param name="baseItem">The base item.</param>
        public Group(ActiveDirectoryGroup baseItem)
        {
            _BaseItem = baseItem;
        }

        /// <summary>
        /// Gets the members of the group.
        /// </summary>
        /// <value>The members of the group.</value>
        public IList<string> Members
        {
            get { return _BaseItem.Members; }
        }

        /// <summary>
        /// Gets the group scope.
        /// </summary>
        /// <value>The group scope.</value>
        public GroupScope? Scope
        {
            get { return _BaseItem.Scope; }
        }

        /// <summary>
        /// Gets the type of the group.
        /// </summary>
        /// <value>The type of the group.</value>
        public GroupType? Type
        {
            get { return _BaseItem.Type; }
        }

        /// <summary>
        /// Gets the name of the account.
        /// </summary>
        /// <value>The name of the account.</value>
        public string NT4AccountName
        {
            get { return _BaseItem.NT4AccountName; }
        }

        /// <summary>
        /// Gets the type of the principal.
        /// </summary>
        /// <value>The type of the principal.</value>
        public PrincipalType? PrincipalType
        {
            get { return _BaseItem.PrincipalType; }
        }

        /// <summary>
        /// Gets the async operation operation used to refresh the item.
        /// </summary>
        /// <returns>An async operation which can be used to refresh the item.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
        public AsyncOperation<RequestProgress> GetRefreshOperation()
        {
            return _BaseItem.GetRefreshOperation();
        }

        /// <summary>
        /// Synchronously refreshes the properties of the curren item.
        /// </summary>
        public void Refresh()
        {
            _BaseItem.Refresh();
        }

        /// <summary>
        /// Asynchronously refreshes the properties of the curren item.
        /// </summary>
        public void RefreshAsync()
        {
            _BaseItem.RefreshAsync();
        }

        /// <summary>
        /// Gets a list containing the properties to load during a refresh operation.
        /// If left empty, all properties are loaded.
        /// </summary>
        /// <value>The properties to load.</value>
        public IList<string> PropertiesToLoad
        {
            get { return _BaseItem.PropertiesToLoad; }
        }

        /// <summary>
        /// Provides access to the properties of the base item.
        /// </summary>
        /// <value>The properties.</value>
        public DirectoryServices.PropertyCollection Properties
        {
            get { return _BaseItem.Properties; }
        }

        /// <summary>
        /// Gets the distinguished name.
        /// </summary>
        /// <value>The distinguished name.</value>
        public string DistinguishedName
        {
            get { return _BaseItem.DistinguishedName; }
        }

        /// <summary>
        /// Gets or sets the displayname of the item.
        /// </summary>
        /// <value>The displayname of the item.</value>
        public string DisplayName
        {
            get { return _BaseItem.DisplayName; }
            set { _BaseItem.DisplayName = value; }
        }

        #region IAddressListEntry Members

        /// <summary>
        /// Gets the name of the entry
        /// </summary>
        /// <value>The name.</value>
        string IAddressListEntry.Name
        {
            get { return DisplayName; }
        }

        #endregion
    }
}
