using System;
using System.Collections.Generic;
using System.Net;
using System.Security.Principal;
using System.Text;
using InfiniTec.Exchange.Properties;
using InfiniTec.Exchange.Security;
using InfiniTec.Exchange.WebDav.Request;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    class CalendarFolder: Folder<CalendarItem>
    {
        private const string LocalFreeBusyPath = "NON_IPM_SUBTREE/FreeBusy Data";
        private Folder _FreeBusyFolder;
        private bool _SecurityHasChanged;

        public CalendarFolder(string path, Uri mailboxUrl, Connection connection) : base(path, connection)
        {
            string freeBusyPath;

            freeBusyPath = mailboxUrl.ToString();
            if (!freeBusyPath.EndsWith("/")) freeBusyPath += "/";
            freeBusyPath += LocalFreeBusyPath;

            _FreeBusyFolder = new Folder(freeBusyPath, connection);
            _FreeBusyFolder.DefaultPropertiesToLoad = DefaultProperties.SecurityInformation;
        }

        public override AsyncOperation<RequestStatus> GetRefreshOperation()
        {
            return new AsyncOperation<RequestStatus>(RefreshInternal(base.GetRefreshOperation()));
        }

        private IEnumerable<OperationAction> RefreshInternal(AsyncOperation<RequestStatus> baseOperation)
        {
            yield return new NestedOperation(baseOperation, true);

            using (AsyncOperation operation = _FreeBusyFolder.GetRefreshOperation())
            {
                yield return new NestedOperation(operation, true, typeof (WebException));

                if (operation.Status == OperationStatus.Error)
                {
                    WebException ex = operation.Exception as WebException;

                    if (ex != null && ((HttpWebResponse) ex.Response).StatusCode == HttpStatusCode.NotFound)
                    {
                        throw new LocalFreeBusyInformationNotFoundException(
                            Resources.LocalFreeBusyInformationNotFoundException, ex);
                    }
                    else
                    {
                        throw operation.Exception;
                    }
                }
            }
        }

        public override void SetSecurity(FolderSecurity security)
        {
            base.SetSecurity(security);
            UpdateFreeBusySecurityDescriptor(security);
            _SecurityHasChanged = true;
        }

        private void UpdateFreeBusySecurityDescriptor(FolderSecurity security)
        {
            List<SecurityIdentifier> list;
            FolderSecurity freeBusyDescriptor;

            freeBusyDescriptor = _FreeBusyFolder.GetSecurity();
            list = new List<SecurityIdentifier>();

            foreach (FolderAce ace in freeBusyDescriptor.GetAccessControlEntries())
            {
                if (!list.Contains(ace.IdentityReference)) list.Add(ace.IdentityReference);
            }

            foreach (FolderAce ace in security.GetAccessControlEntries())
            {
                if (list.Contains(ace.IdentityReference))
                {
                    list.Remove(ace.IdentityReference);
                }

                freeBusyDescriptor.DangerousSetAccess(ace.IdentityReference, 
                    ace.FolderAccessMask, ace.ItemAccessMask);
            }
            foreach (SecurityIdentifier sid in list)
            {
                freeBusyDescriptor.RemoveAccess(sid);
            }

            _FreeBusyFolder.SetSecurity(freeBusyDescriptor);
        }

        public override AsyncOperation<RequestStatus> GetSaveOperation()
        {
            return new AsyncOperation<RequestStatus>(SaveInternal(base.GetSaveOperation()));
        }

        private IEnumerable<OperationAction> SaveInternal(AsyncOperation<RequestStatus> baseOperation)
        {
            yield return new NestedOperation(baseOperation, true);

            if (_SecurityHasChanged)
            {
                yield return new NestedOperation(_FreeBusyFolder.GetSaveOperation(), true);
                _SecurityHasChanged = false;
            }
        }

    }
}
