using System;
using System.Collections.Generic;
using InfiniTec.Exchange.Properties;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// Provides read-only access to distribution lists on an exchange folder.
    /// </summary>
    /// <remarks>
    /// Currently, members of the distribution list can only be resolved if
    /// they reside in the same store (either the same mailbox or the default public 
    /// folder tree).
    /// </remarks>
    [SchemaClass(typeof(WellknownProperties.DistributionList))]
    public class DistributionList: Item, IAddressListEntry
    {
        private static string[] _RequiredFields = SchemaHelper.ExtractFieldValues(typeof(WellknownProperties.DistributionList));
        private IList<IAddressListEntry> _Members;
        private int _UnresolvableEntries;
        private bool _ResolveMembers;

        /// <summary>
        /// Gets the number of unresolvable entries.
        /// </summary>
        /// <value>The number of unresolvable entries.</value>
        public int UnresolvableEntries
        {
            get { return _UnresolvableEntries; }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1811:AvoidUncalledPrivateCode")]
        internal bool ResolveMembers
        {
            get { return _ResolveMembers; }
            set { _ResolveMembers = value; }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="MessageItem"/> class.
        /// </summary>
        public DistributionList()
        {
            Initialize();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="WellknownProperties.Message"/> class.
        /// </summary>
        /// <param name="path">The path of the message.</param>
        /// <param name="connection">The connection.</param>
        public DistributionList(string path, Connection connection)
            : base(path, connection)
        {
            Initialize();
        }

        private void Initialize()
        {
            _ResolveMembers = true;
            AddRequiredFields(_RequiredFields);
            SetContentClass("IPM.DistList", "urn:content-classes:group");
        }

        /// <summary>
        /// Gets the members of distribution list.
        /// </summary>
        /// <value>The members of the distribution list</value>
        /// <exception cref="NotInitializedException">Thrown, when <see cref="Item.Refresh"/> operation has not been called.</exception>
        public IList<IAddressListEntry> Members
        {
            get
            {
                if (_Members == null) throw new NotInitializedException(Resources.InstanceNotInitializedException);

                return _Members;
            }
        }

        #region IAddressListEntry Members

        /// <summary>
        /// Gets the name of the entry
        /// </summary>
        /// <value>The name.</value>
        string IAddressListEntry.Name
        {
            get
            {
                return Subject;
            }
        }

        #endregion

        private List<EntryIdReference> DecodeMembers(Guid ownStoreId)
        {
            MultiValuedProperty<byte[]> memberProperty;
            List<EntryIdReference> result;

            if (!BaseItem.Properties.TryGetMultiValuedProperty(WellknownProperties.DistributionList.Members, out memberProperty))
            {
                throw new NotInitializedException(Resources.InstanceNotInitializedException);
            }

            result = new List<EntryIdReference>();

            foreach (byte[] buffer in memberProperty.Values)
            {
                EntryIdReference reference;
                reference = EntryIdReference.DecodeEntryId(buffer, ownStoreId, GetBaseUrl());
                result.Add(reference);
            }

            return result;
        }

        private string GetBaseUrl()
        {
            Uri url;
            string path;

            url = new Uri(BaseItem.Properties.GetProperty<string>(
                              WellknownProperties.DistributionList.PermanentUrl).Value);
            path = url.PathAndQuery;
            path = path.Substring(0, path.IndexOf("/-FlatUrlSpace-/"));

            return path;
        }

        /// <summary>
        /// Gets the <see cref="AsyncOperation"/> which can be used to refresh
        /// the item.
        /// </summary>
        /// <returns>
        /// A <see cref="AsyncOperation"/> instance.
        /// </returns>
        public override AsyncOperation GetRefreshOperation()
        {
            return new AsyncOperation(RefreshInternal(base.GetRefreshOperation()));
        }

        private IEnumerable<OperationAction> RefreshInternal(AsyncOperation baseOperation)
        {
            Dictionary<NestedOperation,EntryIdReference> operations;
            ExchangeStoreReference entryId;
            List<IAddressListEntry> result;

            _UnresolvableEntries = 0;

            yield return new NestedOperation(baseOperation, false);


            if (!_ResolveMembers)
            {
                _ResolveMembers = true;
                yield break;
            }

            entryId = new ExchangeStoreReference(Properties.GetProperty<byte[]>(WellknownProperties.Item.EntryId).Value);

            operations = new Dictionary<NestedOperation, EntryIdReference>();

            foreach (EntryIdReference reference in DecodeMembers(entryId.StoreId))
            {
                operations.Add(
                    new NestedOperation(reference.GetResolveOperation(BaseItem.Connection), false),
                    reference);
            }

            yield return new ParallelOperations(operations.Keys, 5);

            result = new List<IAddressListEntry>();

            foreach (KeyValuePair<NestedOperation, EntryIdReference> entry in operations)
            {
                AsyncOperation<IAddressListEntry> operation;

                operation = (AsyncOperation<IAddressListEntry>) entry.Key.Operation;
                if (operation.StatusInformation != null)
                {
                    result.Add(operation.StatusInformation);
                }
                else
                {
                    _UnresolvableEntries++;
                }
            }
            
            _Members = result.AsReadOnly();
        }
    }
}
