using System;
using System.Collections.Generic;
using System.Runtime.CompilerServices;
using InfiniTec.Exchange.Properties;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class represents a public or private folder on an Exchange store.
    /// </summary>
    public class Folder : FolderBase, IDisposable
    {
        private ItemCollection _ChildItems;

        /// <summary>
        /// Initializes a new instance of the <see cref="Folder"/> class.
        /// </summary>
        /// <param name="path">The path of the folder on the server.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to retrieve. If omitted, all properties are retrieved.</param>
        public Folder(string path, Connection connection, params string[] propertiesToLoad) :
            base(path, connection, propertiesToLoad)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Folder"/> class.
        /// </summary>
        /// <param name="path">The path.</param>
        /// <param name="connection">The connection.</param>
        /// <param name="isNew">if set to <c>true</c> [is new].</param>
        /// <param name="propertiesToLoad">The properties to load.</param>
        protected Folder(string path, Connection connection, bool isNew, params string[] propertiesToLoad) :
            base(path, connection, isNew, propertiesToLoad)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="FolderBase"/> class.
        /// </summary>
        /// <param name="url">The path of the folder on the server.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="properties">The properties of the element.</param>        
        protected internal Folder(Uri url, Connection connection, IEnumerable<Property> properties) :
            base(url, connection, properties)
        {
        }


        /// <summary>
        /// Returns a <see cref="Searcher"/> which can be used
        /// to search the current folder.
        /// </summary>
        /// <param name="propertiesToLoad">Optional. Specifies the
        /// properties to search for.</param>
        /// <returns>A <see cref="Searcher"/> instance.</returns>
        public Searcher Search(params string[] propertiesToLoad)
        {
            return new Searcher(Url, Connection, propertiesToLoad);
        }

        /// <summary>
        /// Creates a new item with the specified name in this folder.
        /// </summary>
        /// <param name="name">The name of the new item.</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA2234:PassSystemUriObjectsInsteadOfStrings")]
        public SimpleItem Add(string name)
        {
            Uri url;

            if (string.IsNullOrEmpty(name)) throw new ArgumentNullException("name");

            name = name.Normalize();

            if (name.Contains("/")) throw new InvalidNameException(Resources.InvalidFilenameSpecified);

            url = new Uri(Url, name);

            return new SimpleItem(url, Connection, null);
        }

        /// <summary>
        /// Creates a new folder with the specified path.
        /// </summary>
        /// <param name="path">The path of the new folder.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <returns>A <see cref="Folder"/> object</returns>
        public static Folder Create(string path, Connection connection)
        {
            return new Folder(path, connection, true);
        }

        /// <summary>
        /// Creates a new folder with the specified path.
        /// </summary>
        /// <param name="url">The <see cref="Uri"/> of the new folder.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <returns>A <see cref="Folder"/> object</returns>
        public static Folder Create(Uri url, Connection connection)
        {
            Folder result;
            
            result = new Folder(url, connection, null);
            result.FolderNeedsCreation = true;

            return result;
        }

        /// <summary>
        /// Gets a <see cref="ItemCollection"/> instance which provides access to the child folders of this instance.
        /// </summary>
        /// <value>The sub folders.</value>
        public ItemCollection ChildItems
        {
            [MethodImpl(MethodImplOptions.Synchronized)]
            get
            {
                if (_ChildItems == null) _ChildItems = new ItemCollection(Url, Connection);
                return _ChildItems;
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <param name="disposing"><c>true</c> if called from the method. False, otherwise.</param>
        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);

            if (!disposing) return;

            if (_ChildItems != null) _ChildItems.Dispose();
        }
    }

}