using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Globalization;
using InfiniTec.Exchange.Properties;
using InfiniTec.Threading;
using AsyncOperation=InfiniTec.Threading.AsyncOperation;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// Provides access to the child folders in a folder.
    /// </summary>
    public class FolderCollection: ObjectBase, IEnumerable<Folder>, IDisposable
    {
        /// <summary>
        /// This event is called after an <see cref="RefreshAsync"/> operation.
        /// </summary>
        public event AsyncCompletedEventHandler RefreshCompleted;

        private Searcher _Searcher;
        private List<Folder> _Items = new List<Folder>();
        private AsyncOperationContext _AsyncOperationContext = new AsyncOperationContext();


        /// <summary>
        /// Initializes a new instance of the <see cref="ItemCollection&lt;ItemType&gt;"/> class.
        /// </summary>
        /// <param name="url">The url of the element collection.</param>
        /// <param name="connection">The connection.</param>
        internal FolderCollection(Uri url, Connection connection) : base(url, connection)
        {
            Initialize();
        }

        private void Initialize()
        {
            _Searcher = new Searcher(Url, Connection);
            _Searcher.Constraints = string.Format(CultureInfo.InvariantCulture, "(\"{0}\" = false) ", WellknownProperties.Dav.IsFolder);
        }

        /// <summary>
        /// Gets the number of items in this collection
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get
            {
                return _Items.Count;
            }
        }

        /// <summary>
        /// Gets the <see cref="Folder"/> at the specified index.
        /// </summary>
        /// <value></value>
        public Folder this[int index]
        {
           get
           {
               return _Items[index];
           } 
        }

        /// <summary>
        /// Gets or sets the scope of the search.
        /// </summary>
        /// <value>The scope.</value>
        public SearchScope Scope
        {
            get { return _Searcher.Scope; }
            set { _Searcher.Scope = value; }
        }

        /// <summary>
        /// Gets or sets the constraints.
        /// </summary>
        /// <value>The constraints.</value>
        public string Constraints
        {
            get { return _Searcher.Constraints; }
            set { _Searcher.Constraints = value; }
        }

        /// <summary>
        /// Gets or sets the sort order.
        /// </summary>
        /// <value>The sort order</value>
        public string OrderBy
        {
            get { return _Searcher.OrderBy; }
            set { _Searcher.OrderBy = value; }
        }

        IEnumerator<Folder> IEnumerable<Folder>.GetEnumerator()
        {
            return _Items.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator"></see> object that can be used to iterate through the collection.
        /// </returns>
        public IEnumerator GetEnumerator()
        {
            return _Items.GetEnumerator();
        }

        /// <summary>
        /// Synchronously refreshes the properties of the curren item.
        /// </summary>
        public void Refresh()
        {
            using (AsyncOperation operation = GetRefreshOperation())
            {
                operation.RunSynchronously();
            }
        }

        /// <summary>
        /// Asynchronously refreshes the properties of this instance.
        /// </summary>
        public void RefreshAsync()
        {
            AsyncOperation operation;

            _AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetRefreshOperation();

            _AsyncOperationContext.RegisterPendingAsyncOperation(operation);
            operation.Completed += RefreshAsyncCompleted;
            operation.Error += RefreshAsyncError;

            operation.Run();
        }

        /// <summary>
        /// Gets the <see cref="AsyncOperation"/> which can be used to refresh
        /// the item.
        /// </summary>
        /// <returns>A <see cref="AsyncOperation" /> instance.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
        public virtual AsyncOperation GetRefreshOperation()
        {
            return new AsyncOperation(RefreshInternal());
        }

        private IEnumerable<OperationAction> RefreshInternal()
        {
            using (AsyncOperation operation = _Searcher.GetSearchOperation(SearchMode.FindAll))
            {
                yield return new NestedOperation(operation, true);

                _Items.Clear();
                foreach (SimpleItem item in _Searcher.SearchResult)
                {
                    _Items.Add(item.ToFolder());
                }
            }

        }

        private void RefreshAsyncError(object sender, ExceptionEventArgs e)
        {
            _AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnRefreshCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        private void RefreshAsyncCompleted(object sender, CompletedEventArgs e)
        {
            _AsyncOperationContext.UnregisterPendingAsyncOperation();

            OnRefreshCompleted(new AsyncCompletedEventArgs(null, e.Status == OperationStatus.Aborted, null));
        }

        /// <summary>
        /// This method is called after an async refresh operation has completed.
        /// </summary>
        /// <param name="e">An <see cref="AsyncCompletedEventArgs"/> instance indicating the success
        /// or failure of the refresh operation.</param>
        protected virtual void OnRefreshCompleted(AsyncCompletedEventArgs e)
        {
            if (RefreshCompleted != null)
            {
                RefreshCompleted(this, e);
            }
        }

        /// <summary>
        /// Creates a new item with the specified name in this folder.
        /// </summary>
        /// <param name="name">The name of the new item.</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA2234:PassSystemUriObjectsInsteadOfStrings")]
        public Folder Add(string name)
        {
            string url;
            Folder result;

            if (string.IsNullOrEmpty(name)) throw new ArgumentNullException("name");

            name = name.Normalize();

            if (name.Contains("/")) throw new InvalidNameException(Resources.InvalidFilenameSpecified);

            url = Url.ToString();
            if (!url.EndsWith("/")) url += "/";
            url += name;

            result = Folder.Create(url, Connection);

            _Items.Add(result);

            return result;
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <param name="disposing"><c>true</c> if called from the <see cref="ObjectBase.Dispose()"/> method. False, otherwise.</param>
        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);

            if (!disposing) return;

            _AsyncOperationContext.Dispose();
            _Searcher.Dispose();
        }
    }
}
