using System;
using System.Collections.Generic;
using System.Diagnostics;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class performs a search of an Exchange public or private
    /// folder.
    /// </summary>
    public class Searcher : SearcherBase
    {
        /// <summary>
        /// This event is invoked when the progress of the search operation
        /// changed.
        /// </summary>
        public event EventHandler<SearchProgressChangedEventArgs> ProgressChanged;

        private List<SimpleItem> _SearchResult;

        /// <summary>
        /// Gets the search result.
        /// </summary>
        /// <value>The search result.</value>
        public IList<SimpleItem> SearchResult
        {
            [DebuggerStepThrough]
            get { return _SearchResult.AsReadOnly(); }
        }

        /// <summary>
        /// Calls the <see cref="ProgressChanged"/> event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnProgressChanged(SearchProgressChangedEventArgs e)
        {
            if (ProgressChanged != null) ProgressChanged(this, e);
        }

        /// <summary>
        /// Returns the search operation used to perform the search based
        /// on the specified settings.
        /// </summary>
        /// <param name="searchMode">A valid <see cref="SearchMode"/> value.</param>
        /// <returns></returns>
        protected override AsyncOperation GetInternalSearchOperation(SearchMode searchMode)
        {
            return new AsyncOperation<SearchProgress>(SearchOperationInternal(base.GetInternalSearchOperation(searchMode)));
        }

        private IEnumerable<OperationAction> SearchOperationInternal(AsyncOperation baseOperation)
        {
            _SearchResult = new List<SimpleItem>();

            using (baseOperation)
            {
                yield return new NestedOperation(baseOperation, true);
            }
        }

        private void SearchProgressChanged(object sender, ProgressChangedEventArgs<SearchProgress> e)
        {
            SearchProgressChangedEventArgs args;

            args = new SearchProgressChangedEventArgs(e.Status.Range, e.Status.Items);
            OnProgressChanged(args);

            e.Cancel = args.Cancel;
        }

        /// <summary>
        /// Allows the implementors to register events on the async operation which is about to execute.
        /// </summary>
        /// <param name="operation">The operation.</param>
        protected override void RegisterEvents(AsyncOperation operation)
        {
            ((AsyncOperation<SearchProgress>) operation).ProgressChanged += SearchProgressChanged;
        }

        /// <summary>
        /// Gets the asynchronous operation which is used to search the configured search roots.
        /// </summary>
        /// <param name="searchMode">A valid <see cref="SearchMode"/> value.</param>
        /// <returns></returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures"), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures"), System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
        public virtual AsyncOperation<SearchProgress> GetSearchOperation(SearchMode searchMode)
        {
            return (AsyncOperation<SearchProgress>) GetInternalSearchOperation(searchMode);
        }

        /// <summary>
        /// Processes the partial resultset from the last search operation
        /// </summary>
        /// <param name="items">The items of the last search operation</param>
        protected override void ProcessPartialResult(IList<SimpleItem> items)
        {
            _SearchResult.AddRange(items);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Searcher"/> class.
        /// </summary>
        /// <param name="searchRoot">The root of the search operation.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to load. This parameter is optional.</param>
        public Searcher(string searchRoot, Connection connection, params string[] propertiesToLoad)
            :
                base(searchRoot, connection, propertiesToLoad)
        {
            _SearchResult = new List<SimpleItem>();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Searcher"/> class.
        /// </summary>
        /// <param name="searchRoot">The search root.</param>
        /// <param name="connection">The connection.</param>
        /// <param name="propertiesToLoad">The properties to load.</param>
        public Searcher(Uri searchRoot, Connection connection, params string[] propertiesToLoad)
            :
                base(searchRoot, connection, propertiesToLoad)
        {
            _SearchResult = new List<SimpleItem>();
        }
    }
}