using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using InfiniTec.Exchange.WebDav.Request;
using InfiniTec.Threading;
using AsyncOperation=InfiniTec.Threading.AsyncOperation;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class performs a search of an Exchange public or private
    /// folder.
    /// </summary>
    public abstract class SearcherBase : ObjectBase
    {
        /// <summary>
        /// This event is invoked when the the search operation has completed.
        /// </summary>
        public event AsyncCompletedEventHandler Completed;

        private int _PageSize;
        private IList<string> _PropertiesToLoad;
        private string _OrderBy;
        private string _Constraints;
        private SearchScope _Scope;
        private List<Uri> _AdditionalSearchRoots;
        private int _CurrentPage;
        private bool _SearchCompleted;


        /// <summary>
        /// Gets the current page within a paged search operation.
        /// </summary>
        /// <value>The current page.</value>
        public int CurrentPage
        {
            get { return _CurrentPage; }
        }

        /// <summary>
        /// Gets a value indicating whether a paged search has completed.
        /// </summary>
        /// <value><c>true</c> if the search has completed. Otherwise, <c>false</c>.</value>
        public bool SearchCompleted
        {
            get { return _SearchCompleted; }
        }

        /// <summary>
        /// This list contains additional search roots which should be searched along with
        /// the base url.
        /// </summary>
        public IList<Uri> AdditionalSearchRoots
        {
            [DebuggerStepThrough]
            get { return _AdditionalSearchRoots; }
        }

        /// <summary>
        /// Gets or sets the <see cref="SearchScope"/> of the search operation.
        /// </summary>
        public SearchScope Scope
        {
            get { return _Scope; }
            set { _Scope = value; }
        }

        /// <summary>
        /// Gets or sets the constraints of the operation.
        /// </summary>
        /// <remarks>
        /// The constraints is something like
        /// <code>"DAV:ishidden" = false</code>
        /// </remarks>
        public string Constraints
        {
            get { return _Constraints; }
            set { _Constraints = value; }
        }

        /// <summary>
        /// Gets or sets the sorting key of the search.
        /// </summary>
        /// <remarks>The search order can be something like
        /// <code>"DAV:displayname" ASC</code>
        /// </remarks>
        public string OrderBy
        {
            get { return _OrderBy; }
            set { _OrderBy = value; }
        }

        /// <summary>
        /// Provides access to a collection containing the properties to search for.
        /// If this collection is empty, all properties are returned.
        /// </summary>
        public IList<string> PropertiesToLoad
        {
            get { return _PropertiesToLoad; }
        }

        /// <summary>
        /// Gets or sets the number of items returned by each iteration.
        /// </summary>
        public int PageSize
        {
            get { return _PageSize; }
            set { _PageSize = value; }
        }

        /// <summary>
        /// Calls the <see cref="Completed"/> event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnCompleted(AsyncCompletedEventArgs e)
        {
            if (Completed != null) Completed(this, e);
        }

        /// <summary>
        /// Synchronously starts the search operation.
        /// </summary>
        /// <param name="searchMode">A valid <see cref="SearchMode"/> value.</param>
        protected virtual void FindInternal(SearchMode searchMode)
        {
            using (AsyncOperation op = GetInternalSearchOperation(searchMode))
            {
                op.RunSynchronously();
            }
        }

        /// <summary>
        /// Synchronously starts the search operation.
        /// </summary>
        /// <returns></returns>
        public void FindAll()
        {
            FindInternal(SearchMode.FindAll);
        }

        /// <summary>
        /// Synchronously starts the search operation.
        /// </summary>
        /// <returns></returns>
        public void FindPage()
        {
            FindInternal(SearchMode.FindPage);
        }

        /// <summary>
        /// Asynchronously starts the search operation.
        /// </summary>
        private void FindAsyncInternal(SearchMode searchMode)
        {
            AsyncOperation operation;

            AsyncOperationContext.InitiatePendingAsyncOperation();

            operation = GetInternalSearchOperation(searchMode);

            AsyncOperationContext.RegisterPendingAsyncOperation(operation);

            operation.Completed += SearchAsyncCompleted;
            operation.Error += SearchAsyncError;

            RegisterEvents(operation);

            operation.Run();
        }

        /// <summary>
        /// Synchronously starts the search operation.
        /// </summary>
        /// <returns></returns>
        public void FindAllAsync()
        {
            FindAsyncInternal(SearchMode.FindAll);
        }

        /// <summary>
        /// Synchronously starts the search operation.
        /// </summary>
        /// <returns></returns>
        public void FindPageAsync()
        {
            FindAsyncInternal(SearchMode.FindPage);
        }

        /// <summary>
        /// Allows implementors to register events on the async operation which is about to execute.
        /// </summary>
        /// <param name="operation">The operation.</param>
        protected abstract void RegisterEvents(AsyncOperation operation);

        private void SearchAsyncError(object sender, ExceptionEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();
            OnCompleted(new AsyncCompletedEventArgs(e.Exception, false, null));
        }

        private void SearchAsyncCompleted(object sender, CompletedEventArgs e)
        {
            AsyncOperationContext.UnregisterPendingAsyncOperation();

            if (e.Status == OperationStatus.Completed)
                OnCompleted(new AsyncCompletedEventArgs(null, false, null));
            else
                OnCompleted(new AsyncCompletedEventArgs(null, true, null));
        }

        /// <summary>
        /// Returns the search operation used to perform the search based
        /// on the specified settings.
        /// </summary>
        /// <param name="searchMode">A valid <see cref="SearchMode"/> value.</param>
        /// <returns></returns>
        protected virtual AsyncOperation GetInternalSearchOperation(SearchMode searchMode)
        {
            return new AsyncOperation<SearchProgress>(SearchOperationInternal(searchMode));
        }

        private IEnumerable<OperationAction> SearchOperationInternal(SearchMode mode)
        {
            SearchSettings settings;
            int itemsFound;
            SearchProgress progress;
            ProgressUpdate<SearchProgress> update;

            progress = new SearchProgress();
            update = new ProgressUpdate<SearchProgress>(progress);

            settings = new SearchSettings(_PropertiesToLoad);
            settings.Constraints = Constraints;
            settings.OrderBy = OrderBy;
            settings.Scope = Scope;
            foreach (Uri url in AdditionalSearchRoots)
            {
                settings.AdditionalSearchRoots.Add(url);
            }

            do
            {
                settings.Ranges.Clear();
                settings.Ranges.Add(new Range(_CurrentPage * PageSize, (_CurrentPage+1) * PageSize - 1));

                using (AsyncOperation<RequestStatus, SearchResult> op = Methods.Search(Url, settings))
                {
                    yield return new NestedOperation(op, true);

                    itemsFound = op.Result.Items.Count;

                    if (itemsFound > 0)
                    {
                        progress.Range = new Range(_CurrentPage*PageSize, _CurrentPage*PageSize + itemsFound - 1);
                    }
                    else
                    {
                        progress.Range = null;
                    }
                    progress.Items = op.Result.Items;

                    ProcessPartialResult(op.Result.Items);

                    yield return update;

                    if (settings.Ranges[0].To == op.Result.TotalItems - 1)
                    {
                        _SearchCompleted = true;
                        yield break;
                    }
                }

                _CurrentPage++;
            } while (mode == SearchMode.FindAll && itemsFound == PageSize);

            _SearchCompleted = itemsFound < PageSize;
        }

        /// <summary>
        /// Processes the partial resultset from the last search operation
        /// </summary>
        /// <param name="items">The items of the last search operation</param>
        protected abstract void ProcessPartialResult(IList<SimpleItem> items);

        /// <summary>
        /// Creates a new instance of this class using the specified settings and
        /// a PageSize of 200.
        /// </summary>
        /// <param name="searchRoot">The root of the search operation.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to load. This parameter is optional.</param>
        protected SearcherBase(string searchRoot, Connection connection, params string[] propertiesToLoad)
            :
                base(searchRoot, connection)
        {
            Initialize(propertiesToLoad);
        }


        internal SearcherBase(Uri searchRoot, Connection connection, string[] propertiesToLoad)
            :
                base(searchRoot, connection)
        {
            Initialize(propertiesToLoad);
        }

        private void Initialize(string[] propertiesToLoad)
        {
            _PageSize = 200;
            _PropertiesToLoad = new List<string>(propertiesToLoad);
            _AdditionalSearchRoots = new List<Uri>();
        }
    }
}