using System;
using System.Collections.Generic;
using System.Diagnostics;
using InfiniTec.Threading;

namespace InfiniTec.Exchange.WebDav
{
    /// <summary>
    /// This class performs a search of an Exchange public or private
    /// folder.
    /// </summary>
    /// <typeparam name="ItemType">The type of the messages returned by the searcher</typeparam>
    public class Searcher<ItemType> : SearcherBase
        where ItemType : Item, new()
    {
        /// <summary>
        /// This event is invoked when the progress of the search operation
        /// changed.
        /// </summary>
        public event EventHandler<SearchProgressChangedEventArgs<ItemType>> ProgressChanged;

        private List<ItemType> _SearchResult;

        /// <summary>
        /// Gets the search result.
        /// </summary>
        /// <value>The search result.</value>
        public IList<ItemType> SearchResult
        {
            [DebuggerStepThrough]
            get { return _SearchResult.AsReadOnly(); }
        }

        /// <summary>
        /// Processes the partial resultset from the last search operation
        /// </summary>
        /// <param name="items">The items of the last search operation</param>
        protected override void ProcessPartialResult(IList<SimpleItem> items)
        {
            foreach (SimpleItem item in items)
            {
                ItemType element;
                element = new ItemType();
                element.BaseItem = item;

                _SearchResult.Add(element);
            }
        }


        /// <summary>
        /// Returns the search operation used to perform the search based
        /// on the specified settings.
        /// </summary>
        /// <returns></returns>
        protected override AsyncOperation GetInternalSearchOperation(SearchMode searchMode)
        {
            return new AsyncOperation<SearchProgress<ItemType>>(
                    SearchInternal((AsyncOperation<SearchProgress>) base.GetInternalSearchOperation(searchMode)));
        }

        /// <summary>
        /// Gets the search operation.
        /// </summary>
        /// <param name="searchMode">The search mode.</param>
        /// <returns></returns>
        public AsyncOperation<SearchProgress<ItemType>> GetSearchOperation(SearchMode searchMode)
        {
            return (AsyncOperation<SearchProgress<ItemType>>) GetInternalSearchOperation(searchMode);
        }

        private IEnumerable<OperationAction> SearchInternal(AsyncOperation<SearchProgress> baseOperation)
        {
            _SearchResult = new List<ItemType>();

            using (baseOperation)
            {
                baseOperation.ProgressChanged +=
                    new ProgressChangedEventHandler<SearchProgress>(InternalSearchOperation_ProgressChanged);

                yield return new NestedOperation(baseOperation, true, null);
            }
        }

        private void InternalSearchOperation_ProgressChanged(object sender, ProgressChangedEventArgs<SearchProgress> e)
        {
            SearchProgressChangedEventArgs<ItemType> args;
            List<ItemType> result;

            result = new List<ItemType>();
            foreach (SimpleItem item in e.Status.Items)
            {
                ItemType entry;

                entry = new ItemType();
                entry.BaseItem = item;

                result.Add(entry);
            }
            args = new SearchProgressChangedEventArgs<ItemType>(e.Status.Range, result.AsReadOnly());
            OnProgressChanged(args);
        }

        /// <summary>
        /// Allows the implementors to register events on the async operation which is about to execute.
        /// </summary>
        /// <param name="operation">The operation.</param>
        protected override void RegisterEvents(AsyncOperation operation)
        {
            ((AsyncOperation<SearchProgress<ItemType>>) operation).ProgressChanged += SearchOperation_ProgressChanged;
        }

        private void SearchOperation_ProgressChanged(object sender, ProgressChangedEventArgs<SearchProgress<ItemType>> e)
        {
            SearchProgressChangedEventArgs<ItemType> args;

            args = new SearchProgressChangedEventArgs<ItemType>(e.Status.Range, e.Status.Items);
            OnProgressChanged(args);

            e.Cancel = args.Cancel;
        }

        private void OnProgressChanged(SearchProgressChangedEventArgs<ItemType> e)
        {
            if (ProgressChanged != null) ProgressChanged(this, e);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Searcher"/> class.
        /// </summary>
        /// <param name="searchRoot">The root of the search operation.</param>
        /// <param name="connection">The <see cref="Connection"/> to use.</param>
        /// <param name="propertiesToLoad">The properties to load. This parameter is optional.</param>
        public Searcher(string searchRoot, Connection connection, params string[] propertiesToLoad)
            :
                base(searchRoot, connection, propertiesToLoad)
        {
            _SearchResult = new List<ItemType>();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Searcher"/> class.
        /// </summary>
        /// <param name="searchRoot">The search root.</param>
        /// <param name="connection">The connection.</param>
        /// <param name="propertiesToLoad">The properties to load.</param>
        public Searcher(Uri searchRoot, Connection connection, params string[] propertiesToLoad)
            :
                base(searchRoot, connection, propertiesToLoad)
        {
            _SearchResult = new List<ItemType>();
        }
    }
}